#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <asm/uaccess.h>
#include <linux/smp_lock.h>

#include "shfs.h"
#include "shfs_proc.h"

static int shfs_readlink(struct dentry*, char*, int);
static int shfs_follow_link(struct dentry*, struct nameidata*);

struct inode_operations shfs_symlink_inode_operations = {
	readlink:	shfs_readlink,
	follow_link:	shfs_follow_link,
	setattr:	shfs_notify_change,
};

static int
shfs_readlink(struct dentry *dentry, char *buffer, int bufflen)
{
	struct inode *inode = dentry->d_inode;
	struct super_block *sb = inode->i_sb;
	struct shfs_sb_info *info = sb->u.generic_sbp;
	
	char buf[SHFS_PATH_MAX];
	struct shfs_directory *dir;
	struct shfs_dir_list_node *file;
	char *real_name;
	int res;
	
	if (shfs_get_name(dentry->d_parent, buf) < 0)
		return -ENAMETOOLONG;
	DEBUG("%s\n", buf);

//	lock_kernel();
	if (!shfs_lock(info))
		return -EINTR;
	res = shfs_dcache_get(info, buf, &dir);
	if (res < 0) {
		VERBOSE(" shfs_cache_get failed!\n");
		goto error;
	}

	for (file = dir->head; file != NULL; file = file->next)
		if (strcmp(file->entry.name, dentry->d_name.name) == 0)
			break;
	if (!file) {
		VERBOSE(" file not found in cache!!?!\n");
		res = -1;
		goto error;
	}

	real_name = file->entry.name + strlen(file->entry.name) + 4;
	shfs_unlock(info);
//	unlock_kernel();

	return vfs_readlink(dentry, buffer, bufflen, real_name);
	
error:
	shfs_unlock(info);
//	unlock_kernel();
	return shfs_remove_sigpipe(res);
}

static int
shfs_follow_link(struct dentry *dentry, struct nameidata *nd)
{
	struct inode *inode = dentry->d_inode;
	struct super_block *sb = inode->i_sb;
	struct shfs_sb_info *info = sb->u.generic_sbp;
	
	char buf[SHFS_PATH_MAX];
	struct shfs_directory *dir;
	struct shfs_dir_list_node *file;
	char *real_name;
	int res;
	
	if (shfs_get_name(dentry->d_parent, buf) < 0)
		return -ENAMETOOLONG;
	DEBUG(" %s/%s\n", buf, dentry->d_name.name);

//	lock_kernel();
	if (!shfs_lock(info))
		return -EINTR;
	res = shfs_dcache_get(info, buf, &dir);
	
	if (res < 0) {
		VERBOSE(" shfs_dcache_get failed!\n");
		goto error;
	}

	for (file = dir->head; file != NULL; file = file->next)
		if (strcmp(file->entry.name, dentry->d_name.name) == 0)
			break;
	if (!file) {
		VERBOSE(" file not found in cache!!?!\n");
		res = -1;
		goto error;
	}

	real_name = file->entry.name + strlen(file->entry.name) + 4;
	DEBUG("real name: %s\n", real_name);
	shfs_unlock(info);
//	unlock_kernel();
	return  vfs_follow_link(nd, real_name);
	
error:
	shfs_unlock(info);
//	unlock_kernel();
	return shfs_remove_sigpipe(res);
}

